;-----------------------------------------------------------------------------
;
; ph_.asm: selected ph_.c routines in assembly.  Paul Houle 9/24/2019
; Compiled w/JWASM2.12 -5 -coff (32-bit) or -win64 (64-bit)
;
; Objective is small code size, not performance or strict adherence to X64 ABI
; (eg may call internal routine w/o shadow space, or 16-byte aligned stack).
; Also no Win64 SEH info is generated for any routine.
;
; These routines are simple enough that macros are used, as much as possible,
; to have only one source for both 32 and 64 bit.  This approach discourages
; using r8-r15, which yields a codespace bonus, as opcodes (eg push/pop, or
; those with dword operand size) using r8-r15 are larger.
;
;-----------------------------------------------------------------------------

if64	macro	a64, a32		;return a64 or a32 if 64/32 bit
	ifdef	rax			;rax defined only if 64-bit
	  exitm	<a64>
	endif
	exitm	<a32>
	endm

proc_	macro	name:label, args:vararg	;used to build COMDAT enclosed funcs
	.errnz	@InStr(0, <args>,<uses>), <name: uses not allowed>
_TEXT_&name&	segment	byte flat alias(".text") comdat(1) 'CODE'
name		proc	args
	endm
endp_	macro	name:label
name		endp
_TEXT_&name&	ends
	endm

xax	textequ	if64(rax, eax)
xbx	textequ	if64(rbx, ebx)
xcx	textequ	if64(rcx, ecx)
xdx	textequ	if64(rdx, edx)
xsi	textequ	if64(rsi, esi)
xdi	textequ	if64(rdi, edi)
xbp	textequ	if64(rbp, ebp)
xsp	textequ	if64(rsp, esp)
size_t	textequ if64(qword, dword)
;;xpushf	textequ if64(pushfq, pushfd)
;;xpopf	textequ if64(popfq, popfd)

	;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	; Couldn't get an assembler (eg JWASM 2.12, UASM 2.49) to build what
	; we wanted via proc, so proc is essentially shut off -- arg loading
	; and register push/pop must be done manually.  (We do require
	; the proc to decorate the name and make it public.)
	;
	; Another thing required out of proc is that, in 32-bit mode, ret
	; generates a "retn amount" with the correct 'amount' value (which
	; is determined by argument count).  The 64-bit retn 'amount'
	; is always zero (caller removes stacked args/shadow space).
	;- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	if64(,<.model flat,fastcall>)
	.code
	.listall
	option	stackbase:xsp
	if64(<option prologue:none>)
	if64(<option epilogue:none>)

ifww	macro	ww, wdw			;ww if wchar_t is word, wdw if dword
	  exitm	<ww>			;for windows wchar_t is word
	endm
wchr	textequ	ifww(<word>, <dword>)	;wide char type

	; The asm "rep" versions of string funcs are faster -- 60% of
	; the time of loops generated by MSVC 8 (timed on i5-2500k).

if 0 ;ph_?memcpy() unused -- ph_?memcpy #define'd to ph_?memmove in ph_.h
ph_mcpy	macro	name:label, wchr_byte
name	proc_	pDst:ptr, pSrc:ptr, n:size_t
	push	xdi
	mov	xdi,xcx			;xdi= pDst
	push	xsi
	mov	xsi,xdx			;xsi= pSrc
	mov	xax,xcx
	mov	xcx,if64(r8, [esp+3*4])	;xcx= n
	rep movs wchr_byte ptr [xdi], wchr_byte ptr [xsi]
	pop	xsi
	pop	xdi
	ret
name	endp_
	endm
ph_memcpy	ph_mcpy	byte
ph_wmemcpy	ph_mcpy	wchr
endif

;Note: moving backward (low->hi addr) is slower, so we try not to
ph_mmov	macro	name:label, wchr_byte
name	proc_	pDst:ptr, pSrc:ptr, n:size_t
	push	xdi
	mov	xdi,xcx			;xdi= pDst
	push	xsi
	mov	xsi,xdx			;xsi= pSrc
	mov	xax,xcx
	mov	xcx,if64(r8, [esp+3*4])	;xcx= n
	sub	xdx,xdi			;carry if backward (lo->hi move)
	.if	carry?			;if may need to reverse direction,
	  add	xdx,xcx			;xdx= if > 0, size of region overlap
	  ja	@f			;if overlap reverse direction,
	    lea	xdi,[xdi+xcx*sizeof wchr_byte - sizeof wchr_byte]
	    lea	xsi,[xsi+xcx*sizeof wchr_byte - sizeof wchr_byte]
	    std
	  @@:
	.endif
	rep movs wchr_byte ptr [xdi], wchr_byte ptr [xsi]
	cld
	pop	xsi
	pop	xdi
	ret
name	endp_
	endm
ph_memmove	ph_mmov	byte
ph_wmemmove	ph_mmov	wchr

ph_mset	macro	name:label, wchr_byte
name	proc_	pDst:ptr, chr:wchr_byte, n:size_t
	push	xdi
	mov	xdi,xcx			;xdi= pDst
	push	xcx
	mov	eax,edx			;xax= chr
	mov	xcx,if64(r8, [esp+3*4])	;xcx= n
	rep stos wchr_byte ptr [xdi]
	pop	xax
	pop	xdi
	ret
name	endp_
	endm
ph_memset	ph_mset	byte
ph_wmemset	ph_mset	wchr

ph_slen	macro	name:label, wchr_byte
name	proc_	pSrc:ptr
	push	xdi
	mov	xdi,xcx			;xdi= pSrc
	xor	eax,eax
	or	xcx,-1
	repne scas wchr_byte ptr [xdi]
	sub	xax,xcx
	sub	xax,2
	pop	xdi
	ret
name	endp_
	endm
ph_strlen	ph_slen	byte
ph_wcslen	ph_slen	wchr

ph_scpy	macro	name:label, wchr_byte
name	proc_	pDst:ptr, pSrc:ptr
	push	xdi
	mov	xdi,xcx			;xdi= pDst
	sub	xdi,xdx
	if	wchr_byte eq byte
	  .repeat
	    mov  al,[xdx]
	    mov  [xdi+xdx],al
	    inc  xdx
	  .until !al
	else
	  .repeat
	    ifww(movzx,mov) eax,wchr_byte ptr [xdx]
	    mov   [xdi+xdx],ifww(ax,eax)
	    add   xdx,sizeof wchr_byte
	  .until  !eax
	endif
	mov	xax,xcx
	pop	xdi
	ret
name	endp_
	endm
ph_strcpy	ph_scpy	byte
ph_wstrcpy	ph_scpy	wchr

ph_schr	macro	name:label, wchr_byte
name	proc_	pSrc:ptr, chr:wchr_byte
	if	wchr_byte eq byte
	  .repeat			;xdl= chr
	    mov  al,[xcx]
	    inc	 xcx
	    cmp	 al,dl
	    jz   @f
	  .until !al
	  xor	 eax,eax
	else
	  ifww(<movzx edx,dx>)		;xdx= chr
	  .repeat
	    ifww(movzx,mov) eax,wchr_byte ptr [xcx]
	    add  xcx,sizeof wchr_byte
	    cmp	 eax,edx
	    jz   @f
	  .until !eax
	endif
	ret
@@:	lea	xax,[xcx - sizeof wchr_byte]
	ret
name	endp_
	endm
ph_strchr	ph_schr	byte
ph_wcschr	ph_schr	wchr

	end
